<?php

namespace Drupal\xnttmulti\Plugin\ExternalEntities\StorageClient;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Core\Database\DatabaseExceptionWrapper;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\external_entities\ExternalEntityInterface;
use Drupal\external_entities\ExternalEntityTypeInterface;
use Drupal\external_entities\Plugin\PluginFormTrait;
use Drupal\external_entities\ResponseDecoder\ResponseDecoderFactoryInterface;
use Drupal\external_entities\StorageClient\ExternalEntityStorageClientBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\dbxschema\Database\CrossSchemaConnectionInterface;
use Drupal\xnttmulti\Exception\MultiExternalEntityException;
use Drupal\xnttmulti\Entity\ReadOnlyExternalEntitySubType;
use Symfony\Component\Yaml\Yaml;

/**
 * External entities multiple storage client.
 *
 * @ExternalEntityStorageClient(
 *   id = "xnttmulti",
 *   label = @Translation("Multiple storages"),
 *   description = @Translation("Generates a combined record from multiple storage sources.")
 * )
 */
class Multi extends ExternalEntityStorageClientBase implements PluginFormInterface {

  use PluginFormTrait;

  /**
   * Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * The external storage client manager.
   *
   * @var \Drupal\Component\Plugin\PluginManagerInterface
   */
  protected $storageClientManager;

  /**
   * Constructs a multiple external storage object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   * @param \Drupal\external_entities\ResponseDecoder\ResponseDecoderFactoryInterface $response_decoder_factory
   *   The response decoder factory service.
   * @param \Drupal\Component\Plugin\PluginManagerInterface $storage_client_manager
   *   The external storage client manager.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
   *   Logger service (xnttmulti).
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    TranslationInterface $string_translation,
    ResponseDecoderFactoryInterface $response_decoder_factory,
    PluginManagerInterface $storage_client_manager,
    MessengerInterface $messenger,
    LoggerChannelInterface $logger
  ) {
    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $string_translation,
      $response_decoder_factory
    );
    // Services injection.
    $this->storageClientManager = $storage_client_manager;
    $this->messenger = $messenger;
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('string_translation'),
      $container->get('external_entities.response_decoder_factory'),
      $container->get('plugin.manager.external_entities.storage_client'),
      $container->get('messenger'),
      $container->get('logger.channel.xnttmulti')
    );
  }

  /**
   * Provides the list of available plugins.
   *
   * @param bool $reset
   *   Set to TRUE to reload the list of plugin (clears the static cache).
   *   Default: FALSE (use the static cache).
   *
   * @return array
   *   An array with the following keys containing arrays keyed by plugin ids:
   *   - plugins: an array of storage client plugin objects;
   *   - definitions: an array of plugin descriptions;
   *   - options: an array ready to be used for "radios" form elements.
   */
  protected function getOtherAvailablePlugins(bool $reset = FALSE) {
    static $storage_client_options = ['options' => [], 'plugins' => [], ];
    if (empty($storage_client_options['options']) || $reset) {
      // Get all storage definitions to provide a list of storage plugins.
      $all_storage_clients = $this->storageClientManager->getDefinitions();
      foreach ($all_storage_clients as $storage_client_id => $definition) {
        // Exclude this plugin from the list.
        if ($this->getPluginId() == $storage_client_id) {
          continue;
        }
        $config = [
          '_external_entity_type' => $this->externalEntityType,
        ];

        $storage_client = $this->storageClientManager->createInstance(
          $storage_client_id,
          $config
        );
        $storage_client_options['plugins'][$storage_client_id] =
          $storage_client
        ;
        $storage_client_options['definitions'][$storage_client_id] =
          $definition
        ;
        $storage_client_options['options'][$storage_client_id] =
          $storage_client->getLabel()
        ;
      }
      asort($storage_client_options['options'], SORT_NATURAL | SORT_FLAG_CASE);
      $storage_client_options['options'] =
        ['' => 'Remove']
        + $storage_client_options['options']
      ;
    }
    return $storage_client_options;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'storages' => [],
    ];
  }

  /**
   * Initialize externalEntityType member if not set.
   */
  protected function initExternalEntityType(FormStateInterface $form_state) {
    // Get current external entity type if not set.
    if (empty($this->externalEntityType)) {
      $state_data = $form_state->getStorage();
      $xntt_type_id = $state_data['machine_name.initial_values']['id'];
      if (!empty($xntt_type_id)) {
        $xntt_type = \Drupal::entityTypeManager()
          ->getStorage('external_entity_type')
          ->load($xntt_type_id)
        ;
        $this->externalEntityType = $xntt_type;
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $this->initExternalEntityType($form_state);

    // Get the list of available storage plugins.
    $storage_clients = $this->getOtherAvailablePlugins();

    $form['storages'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Multiple storage settings'),
    ];
    $form['storages']['description'] = [
      '#type' => '#markup',
      '#markup' =>
        '<div>'
        . $this->t(
          'You can specify one or more storage client. Their order is important: the first one is loaded first.'
        )
        . '</div>'
      ,
    ];
    // Get storages from current form or from config.
    $storages =
      $form_state->get('storages')
      ?? $this->configuration['storages']
      ?? []
    ;
    $storage_count = $form_state->get('storage_count') ?? count($storages);
    // We must have at least one storage.
    if (empty($storage_count)) {
      $storage_count = 1;
    }

    $storage_index = 0;
    // Make sure we work on a non-empy list of plugins.
    if (!empty($storage_clients['options'])) {
      for ($i = 0; $i < $storage_count; ++$i) {
        $default_client = $storages[$i]['storage_client'] ?? '';
        // @todo: fix issue when changing plugin: config corresponds to previous
        // plugin instead of the selected one. Maybe check
        // validateConfigurationForm().
        $client_config =
          $storages[$i]['config']
          ?? []
        ;
        
        // @todo: fix form issue "invalid plugin selection".
        // Issue: Drupal form API will auto-check the wrong radiobox on AJAX
        // reload. If we have 3 storage clients and remove the middle one, the
        // last one will become the "second" item but "Remove" radio will be
        // checked while the displayed settings will correspond to the real
        // last plugin and the 3rd one will remain as an empty plugin but
        // without "Remove" being selected.

        // Only display non-empty clients unless it is the last one.
        if ((!empty($default_client)) || ($i == $storage_count - 1)) {
          if (empty($client_config)
              && !empty($default_client)
              && !empty($this->externalEntityType)
              && ($this->externalEntityType instanceof ExternalEntityTypeInterface)
              && ($this->externalEntityType->getStorageClientId() == $default_client)
          ) {
            $client_config = $this->externalEntityType->getStorageClientConfig();
          }

          $form['storages'][$storage_index] = [
            '#type' => 'details',
            '#title' => $this->t(
              'Storage client #@index',
              ['@index' => $storage_index+1]
            ),
            '#open' => TRUE,
          ];
          $form['storages'][$storage_index]['storage_client'] = [
            '#type' => 'radios',
            '#title' => $this->t('Select a storage client.'),
            '#options' => $storage_clients['options'],
            '#default_value' => $default_client,
            // Only require first element.
            '#required' => (0 === $storage_index),
            '#ajax' => [
              'callback' => [
                get_class($form_state->getFormObject()),
                'buildAjaxStorageClientConfigForm',
              ],
              'wrapper' => 'external-entities-storage-client-config-form',
              'method' => 'replace',
              'effect' => 'fade',
            ],
            '#attributes' => ['autocomplete' => 'off',],
          ];

          // Get client config.
          if (!empty($default_client)) {
            // Attach the storage client plugin configuration form.
            $form['storages'][$storage_index]['config'] = [];
            $storage_client = $storage_clients['plugins'][$default_client];
            $storage_client->setConfiguration($client_config);

            $storage_client_form_state = SubformState::createForSubform(
              $form['storages'][$storage_index]['config'],
              $form,
              $form_state
            );
            $form['storages'][$storage_index]['config'] =
              $storage_client->buildConfigurationForm(
                $form['storages'][$storage_index]['config'],
                $storage_client_form_state
              )
            ;
            $form['storages'][$storage_index]['config'] += [
              '#type' => 'fieldset',
              '#title' => $this->t(
                ':plugin Settings',
                [':plugin' => $storage_clients['options'][$default_client],]
              ),
              '#open' => TRUE,
            ];
            $description =
              $storage_clients['definitions'][$default_client]['description']
            ;
            if (!empty($description)) {
              $form['storages'][$storage_index]['config'] +=
                ['#description' => $description]
              ;
            }

            // The use of override or other identifier field can not work with
            // the first storage client as no field has been loaded for it.
            if (0 !== $storage_index) {
              $form['storages'][$storage_index]['override'] = [
                '#type' => 'checkbox',
                '#title' => $this->t('Override previous field values'),
                '#description' => $this->t('If set, entity field values provided by this plugin will override existing values with the same field name provided by previous plugins.'),
                '#default_value' => $storages[$i]['override'] ?? TRUE,
              ];

              $form['storages'][$storage_index]['id_field'] = [
                '#type' => 'textfield',
                '#title' => $this->t('Field name to use as identifier for this storage client.'),
                '#description' => $this->t('Leave empty to use the default id provided. Otherwise, any text/numeric field provided by previous storage plugins can be used here.'),
                '#default_value' => $storages[$i]['id_field'] ?? NULL,
              ];
            }

            $form['storages'][$storage_index]['readonly'] = [
              '#type' => 'checkbox',
              '#title' => $this->t('Read only (no create/update) from client.'),
              '#default_value' => $storages[$i]['readonly'] ?? FALSE,
            ];

          }
          else {
            $form['storages'][$storage_index]['config'] = [
              '#type' => 'container',
            ];
          }
          ++$storage_index;
        }
      }
    }
    $storage_count = $storage_index;
    $form_state->set('storage_count', $storage_count);
    $form['storages']['add_storage'] = [
      '#type' => 'submit',
      '#value' => $this->t('Add a storage'),
      '#name' => 'add_storage',
      '#ajax' => [
        'callback' => [
          get_class($form_state->getFormObject()),
          'buildAjaxStorageClientConfigForm'
        ],
        'wrapper' => 'external-entities-storage-client-config-form',
        'method' => 'replace',
        'effect' => 'fade',
      ],
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   *
   * Note: there is no ::submitConfigurationForm because External Entities
   * plugin system does not use it. All is done here.
   */
  public function validateConfigurationForm(
    array &$form,
    FormStateInterface $form_state
  ) {
    $this->initExternalEntityType($form_state);

    $update_count = FALSE;
    // Check for Ajax events.
    if ($trigger = $form_state->getTriggeringElement()) {
      switch ($trigger['#name']) {
        case 'add_storage':
          $storage_count = $form_state->get('storage_count') + 1;
          $form_state->set('storage_count', $storage_count);
          $form_state->setRebuild(TRUE);
          $update_count = TRUE;
          break;

        default:
      }
    }

    $storages = $form_state->getValue('storages');
    $storage_clients = $this->getOtherAvailablePlugins();
    $storage_client_configs = [];
    foreach ($storages as $storage_index => $storage) {
      // Skip non-client items or empty clients.
      if (!is_int($storage_index) || empty($storage['storage_client'])) {
        continue;
      }
      $storage_client_id = $storage['storage_client'];
      // Get plugin.
      $storage_client = $storage_clients['plugins'][$storage_client_id] ?? NULL;
      if (empty($storage_client)) {
        $update_count = TRUE;
        continue;
      }
      $storage_client_form_state = SubformState::createForSubform(
        $form['storages'][$storage_index]['config'],
        $form,
        $form_state
      );
      $storage_client->validateConfigurationForm(
        $form['storages'][$storage_index]['config'],
        $storage_client_form_state
      );

      $storage_client_configs[] = [
        'storage_client' => $storage_client_id,
        'id_field' => $storage['id_field'] ?? '',
        'override' => $storage['override'] ?? FALSE,
        'readonly' => $storage['readonly'] ?? TRUE,
        'config' => $storage_client->getConfiguration(),
      ];
    }

    // Update storage count according to "Add storage" button or used "Remove"
    // radios.
    if ($update_count) {
      $form_state->set(
        'storage_count',
        $storage_count ?? count($storage_client_configs)
      );
      $form_state->setRebuild(TRUE);
    }

    //@todo: maybe try to set checked radios here for plugin selection to fix
    // issue "invalid plugin selection".
    // Tryed in $form without success.
    //       $form['storages'][$form_index]['storage_client']['#value'] = $storage_client_id;
    //       $form['storages'][$form_index]['storage_client']['#default_value'] = $storage_client_id;
    //       $form['storages'][$form_index]['storage_client'][$storage_client_id]['#default_value'] = $storage_client_id;
    //       $form['storages'][$form_index]['storage_client'][$storage_client_id]['#checked'] = TRUE;

    $form_state->setValue('storages', $storage_client_configs);
    // Save state for form update when adding items.
    $form_state->set('storages', $storage_client_configs);
    $this->setConfiguration($form_state->getValues());
  }

  /**
   * Returns the ordered list of configured plugin for current instance.
   */
  public function getPlugins() {
    // @todo: static cache?
    $plugins = [];
    $storages = $this->getConfiguration()['storages'] ?? [];
    foreach ($storages as $storage_settings) {
      // Sets externalEntityType member if missing in plugin config.
      if (!empty($this->externalEntityType)
          && ($this->externalEntityType instanceof ExternalEntityTypeInterface)
      ) {
        // Create a "read-only" clone of the real external entity type.
        $client_xntt_type = ReadOnlyExternalEntitySubType::createFromParent(
          $this->externalEntityType
        );
        // Redefines id mapping if needed.
        if (!empty($storage_settings['id_field'])) {
          $field_mapper = $client_xntt_type->getFieldMapper();
          $field_mappings = $field_mapper->getFieldMappings();
          $field_mappings['id'] = ['value' => $storage_settings['id_field']];
          // @todo: check if this works properly.
          $client_xntt_type->set('field_mappings', $field_mappings);
        }
        $storage_settings['config'] +=
          ['_external_entity_type' => $client_xntt_type,]
        ;
      }
      $plugins[] = $this->storageClientManager->createInstance(
        $storage_settings['storage_client'],
        $storage_settings['config']
      );
    }

    return $plugins;
  }

  /**
   * {@inheritdoc}
   */
  public function delete(ExternalEntityInterface $entity) {
    $plugins = $this->getPlugins();
    $storage_config = $this->getConfiguration()['storages'];
    $first_plugin = array_shift($plugins);
    $result = $first_plugin->delete($entity);
    foreach ($plugins as $plugin_index => $plugin) {
      if (empty($storage_config[$plugin_index + 1]['readonly'])) {
        $plugin->delete($entity);
      }
    }
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function loadMultiple(array $ids = NULL) {
    $data = [];
    // Load each entity.
    // @todo: load multiples entities by batch and then merge each parts instead
    // of loading them 1 by 1.
    if (!empty($ids) && is_array($ids)) {
      foreach ($ids as $id) {
        $data[$id] = $this->load($id);
      }
    }
    return $data;
  }

  /**
   * Loads one entity.
   *
   * @param mixed $id
   *   The ID of the entity to load.
   *
   * @return array|null
   *   A raw data array, NULL if no data returned.
   */
  public function load($id) {
    $entity = [];
    $plugins = $this->getPlugins();
    $storage_config = $this->getConfiguration()['storages'];
    foreach ($plugins as $plugin_index => $plugin) {
      // Load and merge values from each storage client.
      $sub_id = $id;
      if (!empty($storage_config[$plugin_index]['id_field'])) {
        $sub_id = $entity[$storage_config[$plugin_index]['id_field']];
      }
      if (!empty($sub_id)) {
        $plugin_entity_data = $plugin->load($sub_id) ?? [];
        if (!empty($storage_config[$plugin_index]['override'])
            && $storage_config[$plugin_index]['override']
        ) {
          // Override any field if needed.
          $entity = NestedArray::mergeDeep($entity, $plugin_entity_data);
        }
        else {
          // No override of fields already set.
          $entity = NestedArray::mergeDeep($plugin_entity_data, $entity);
        }
      }
    }
    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  public function save(ExternalEntityInterface $entity) {
    $plugins = $this->getPlugins();
    $storage_config = $this->getConfiguration()['storages'];
    $first_plugin = array_shift($plugins);
    $result = $first_plugin->save($entity);
    foreach ($plugins as $plugin_index => $plugin) {
      if (empty($storage_config[$plugin_index + 1]['readonly'])) {
        $plugin->save($entity);
      }
    }
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function query(
    array $parameters = [],
    array $sorts = [],
    $start = NULL,
    $length = NULL,
    bool $reload = FALSE
  ) {
    $entities = [];
    $plugins = $this->getPlugins();
    // List entities with first plugin only.
    $plugin = $plugins[0];
    if ($plugin) {
      $entities = $plugin->query($parameters, $sorts, $start, $length, $reload);
    }
    return $entities;
  }

  /**
   * {@inheritdoc}
   */
  public function countQuery(array $parameters = []) {
    $count = count($this->query($parameters));
    return $count;
  }

}
